#ifndef COLOR_CC
#define COLOR_CC 1

#include <iostream>
#include <math.h>
#include <stdlib.h>
#include <strings.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

using namespace std;

int compareFloat(float a, float b) {
  return (a == b) ? 0 : ( (a < b) ? -1 : 1);
}

int sortFloat(const void * a, const void * b) {
  return compareFloat(*(const float *)a, *(const float *)b);
}

struct Color {
  float r, g, b;
  bool operator==(Color c2) {
    return (r == c2.r && g == c2.g && b == c2.b);
  };
  Color operator+(Color c2) {
    return (Color) { r + c2.r, g + c2.g, b + c2.b };
  };
  Color operator-(Color c2) {
    return (Color) { r - c2.r, g - c2.g, b - c2.b };
  };
  Color operator*(float a) {
    return (Color) { r*a, g*a, b*a };
  };
  Color operator/(float a) {
    return (Color) { r/a, g/a, b/a };
  };
  Color floor() {
    return (Color) { floorf(r), floorf(g), floorf(b) };
  }
  double distance(Color c2) {
    Color diff = *this - c2;
    return sqrt(diff.r * diff.r + diff.g * diff.g + diff.b * diff.b);
  }
  bool isPrimary(float margin) {
    if (r > g + margin && r > b + margin) {
      return true;
    } else if (g > b + margin && g > r + margin) {
      return true;
    } else if (b > r + margin && b > g + margin) {
      return true;
    } else {
      return false;
    }
  }
  Color round() {
    return (Color) { roundf(r), roundf(g), roundf(b) };
  }
  Color abs() {
    return (Color) { fabsf(r), fabsf(g), fabsf(b) };
  }
  float lightness() const {
    return r + g + b;
  }
  float saturation() {
    float * s = sorted();
    return s[0] == 0 ? 0 : 1 - s[0]/s[2];
  }
  float getComponent(int index) {
    if (index == 0) {
      return r;
    } else if (index == 1) {
      return g;
    } else if (index == 2) {
      return b;
    }
    return 0;
  }
  void setComponent(int index, float value) {
    if (index == 0) {
      r = value;
    } else if (index == 1) {
      g = value;
    } else if (index == 2) {
      b = value;
    }
  }
  float * values() {
    static float s[3];
    s[0] = r;
    s[1] = g;
    s[2] = b;
    return s;
  }
  void getHSV(float * val) {
    float * s = sorted();
    float diff = s[2] - s[0];
    if (diff == 0) {
      val[0] = 0;
    } else if (s[2] == r) {
      val[0] = fmodf(60.0 * (g-b) / diff + 360, 360);
    } else if (s[2] == g) {
      val[0] = 60.0 * (b-r) / diff + 120;
    } else {
      val[0] = 60.0 * (r-g) / diff + 240;
    }
    if (s[2] == 0) {
      val[1] = 0;
    } else {
      val[1] = 1 - s[0] / s[2];
    }
    val[2] = s[2];
  }
  void setByValue(float * s) {
    r = s[0];
    g = s[1];
    b = s[2];
  }
  void getSorted(float * s) {
    s[0] = r;
    s[1] = g;
    s[2] = b;
    qsort(s, 3, sizeof(float), sortFloat);
  }
  float * sorted() {
    float * s = values();
    qsort(s, 3, sizeof(float), sortFloat);
    return s;
  }
  float middle() {
    if ((r >= g && g >= b) || (b >= g && g >= r)) {
      return g;
    } else if ((g >= b && b >= r) || (r >= b && b >= g)) {
      return b;
    } else {
      return r;
    }
  }
  static Color fromImage(XImage * im, int x, int y) {
    Color result;
    ulong pixel = XGetPixel(im, x, y);
    result.r = (float)(pixel & im->red_mask) / im->red_mask;
    result.g = (float)(pixel & im->green_mask) / im->green_mask;
    result.b = (float)(pixel & im->blue_mask) / im->blue_mask;
    return result;
  }
  void toImage(XImage * im, int x, int y) {
    ulong pixel = ((ulong)(r * im->red_mask) & im->red_mask);
    pixel += ((ulong)(g * im->green_mask) & im->green_mask);
    pixel += ((ulong)(b * im->blue_mask) & im->blue_mask);
    XPutPixel(im, x, y, pixel);
  }
  bool match(double * match) {
    return (fabs(r - match[0]) + fabs(g - match[1]) + fabs(b - match[2])) <= match[3];
  }
  friend ostream& operator<<(ostream& os, Color c);
};

ostream& operator<<(ostream& os, Color c)
{
  os << "(" << c.r << ", " << c.g << ", " << c.b << ")";
  return os;
}

struct PaintColour {
  const char * name;
  Color value;
};

PaintColour colours[] = {
  { "Alice Blue", { 240, 248, 255} },
  { "Antique White", { 250, 235, 215} },
  { "Aqua", { 0, 255, 255} },
  { "Aquamarine", { 127, 255, 212} },
  { "Azure", { 240, 255, 255} },
  { "Beige", { 245, 245, 220} },
  { "Bisque", { 255, 228, 196} },
  { "Black", { 1, 1, 1} },
  { "Blanched Almond", { 255, 235, 205} },
  { "Blue Violet", { 138, 43, 226} },
  { "Blue", { 0, 0, 255} },
  { "Brown", { 165, 42, 42} },
  { "Burly Wood", { 222, 184, 135} },
  { "Cadet Blue", { 95, 158, 160} },
  { "Carrot", { 224, 112, 32} },
  { "Chartreuse", { 127, 255, 0} },
  { "Chocolate", { 210, 105, 30} },
  { "Coral", { 255, 127, 80} },
  { "Cornflower Blue", { 100, 149, 237} },
  { "Cornsilk", { 255, 248, 220} },
  { "Crimson", { 220, 20, 60} },
  { "Dark Blue", { 0, 0, 139} },
  { "Dark Cyan", { 0, 139, 139} },
  { "Dark Golden Rod", { 184, 134, 11} },
  { "Dark Gray", { 169, 169, 169} },
  { "Dark Green", { 0, 100, 0} },
  { "Dark Khaki", { 189, 183, 107} },
  { "Dark Magenta", { 139, 0, 139} },
  { "Dark Olive Green", { 85, 107, 47} },
  { "Dark Orange", { 255, 140, 0} },
  { "Dark Orchid", { 153, 50, 204} },
  { "Dark Red", { 139, 0, 0} },
  { "Dark Salmon", { 233, 150, 122} },
  { "Dark Sea Green", { 143, 188, 143} },
  { "Dark Slate Blue", { 72, 61, 139} },
  { "Dark Slate Gray", { 47, 79, 79} },
  { "Dark Turquoise", { 0, 206, 209} },
  { "Dark Violet", { 148, 0, 211} },
  { "Deep Pink", { 255, 20, 147} },
  { "Deep Sky Blue", { 0, 191, 255} },
  { "Dim Gray", { 105, 105, 105} },
  { "Dodger Blue", { 30, 144, 255} },
  { "Feldspar", { 209, 146, 117} },
  { "Fire Brick", { 178, 34, 34} },
  { "Floral White", { 255, 250, 240} },
  { "Forest Green", { 34, 139, 34} },
  { "Fuchsia", { 255, 0, 255} },
  { "Gainsboro", { 220, 220, 220} },
  { "Ghost White", { 248, 248, 255} },
  { "Gold", { 255, 215, 0} },
  { "Golden Rod", { 218, 165, 32} },
  { "Gray", { 128, 128, 128} },
  { "Green Yellow", { 173, 255, 47} },
  { "Green", { 0, 128, 0} },
  { "Honey Dew", { 240, 255, 240} },
  { "Hot Pink", { 255, 105, 180} },
  { "Indian Red", { 205, 92, 92} },
  { "Indigo", { 75, 0, 130} },
  { "Ivory", { 255, 255, 240} },
  { "Khaki", { 240, 230, 140} },
  { "Lavender Blush", { 255, 240, 245} },
  { "Lavender", { 230, 230, 250} },
  { "Lawn Green", { 124, 252, 0} },
  { "Lemon Chiffon", { 255, 250, 205} },
  { "Light Blue", { 173, 216, 230} },
  { "Light Coral", { 240, 128, 128} },
  { "Light Cyan", { 224, 255, 255} },
  { "Light Golden Rod Yellow", { 250, 250, 210} },
  { "Light Green", { 144, 238, 144} },
  { "Light Grey", { 211, 211, 211} },
  { "Light Pink", { 255, 182, 193} },
  { "Light Salmon", { 255, 160, 122} },
  { "Light Sea Green", { 32, 178, 170} },
  { "Light Sky Blue", { 135, 206, 250} },
  { "Light Slate Blue", { 132, 112, 255} },
  { "Light Slate Gray", { 119, 136, 153} },
  { "Light Steel Blue", { 176, 196, 222} },
  { "Light Yellow", { 255, 255, 224} },
  { "Lime Green", { 50, 205, 50} },
  { "Lime", { 0, 255, 0} },
  { "Linen", { 250, 240, 230} },
  { "Maroon", { 128, 0, 0} },
  { "Medium Aqua Marine", { 102, 205, 170} },
  { "Medium Blue", { 0, 0, 205} },
  { "Medium Orchid", { 186, 85, 211} },
  { "Medium Purple", { 147, 112, 216} },
  { "Medium Sea Green", { 60, 179, 113} },
  { "Medium Slate Blue", { 123, 104, 238} },
  { "Medium Spring Green", { 0, 250, 154} },
  { "Medium Turquoise", { 72, 209, 204} },
  { "Medium Violet Red", { 199, 21, 133} },
  { "Midnight Blue", { 25, 25, 112} },
  { "Mint Cream", { 245, 255, 250} },
  { "Misty Rose", { 255, 228, 225} },
  { "Moccasin", { 255, 228, 181} },
  { "Navajo White", { 255, 222, 173} },
  { "Navy", { 0, 0, 128} },
  { "Old Lace", { 253, 245, 230} },
  { "Olive Drab", { 107, 142, 35} },
  { "Olive", { 128, 128, 0} },
  { "Orange Red", { 255, 69, 0} },
  { "Orange", { 255, 165, 0} },
  { "Orchid", { 218, 112, 214} },
  { "Pale Golden Rod", { 238, 232, 170} },
  { "Pale Green", { 152, 251, 152} },
  { "Pale Turquoise", { 175, 238, 238} },
  { "Pale Violet Red", { 216, 112, 147} },
  { "Papaya Whip", { 255, 239, 213} },
  { "Peach Puff", { 255, 218, 185} },
  { "Peru", { 205, 133, 63} },
  { "Pink", { 255, 192, 203} },
  { "Plum", { 221, 160, 221} },
  { "Powder Blue", { 176, 224, 230} },
  { "Purple", { 128, 0, 128} },
  { "Red", { 255, 0, 0} },
  { "Rosy Brown", { 188, 143, 143} },
  { "Royal Blue", { 65, 105, 225} },
  { "Saddle Brown", { 139, 69, 19} },
  { "Salmon", { 250, 128, 114} },
  { "Sandy Brown", { 244, 164, 96} },
  { "Sea Green", { 46, 139, 87} },
  { "Sea Shell", { 255, 245, 238} },
  { "Sienna", { 160, 82, 45} },
  { "Silver", { 192, 192, 192} },
  { "Sky Blue", { 135, 206, 235} },
  { "Slate Blue", { 106, 90, 205} },
  { "Slate Gray", { 112, 128, 144} },
  { "Snow", { 255, 250, 250} },
  { "Spring Green", { 0, 255, 127} },
  { "Steel Blue", { 70, 130, 180} },
  { "Tan", { 210, 180, 140} },
  { "Teal", { 0, 128, 128} },
  { "Thistle", { 216, 191, 216} },
  { "Tomato", { 255, 99, 71} },
  { "Turquoise", { 64, 224, 208} },
  { "Violet Red", { 208, 32, 144} },
  { "Violet", { 238, 130, 238} },
  { "Wheat", { 245, 222, 179} },
  { "White Smoke", { 245, 245, 245} },
  { "White", { 255, 255, 255} },
  { "Yellow Green", { 154, 205, 50} },
  { "Yellow", { 255, 255, 0} },
};

#define numColours (sizeof(colours) / sizeof(PaintColour))

PaintColour * getByName(const char * name)
{
  for (uint i = 0; i < sizeof(colours) / sizeof(PaintColour); i++) {
    if (strcasecmp(name, colours[i].name) == 0) {
      return &colours[i];
    }
  }
  return NULL;
}

PaintColour * nearestColour(Color test)
{
  double minDist = 1000;
  PaintColour * nearest = NULL;
  for (uint i = 0; i < numColours; i++) {
    double dist = test.distance(colours[i].value);
    if (dist < minDist) {
      minDist = dist;
      nearest = & colours[i];
    }
  }
  return nearest;
}

int paintIndex(PaintColour * c) {
  return c - colours;
}

#endif
