#include <X11/Xlib.h>
#include <X11/X.h>
#include <X11/Xutil.h>
#include <X11/extensions/XTest.h>
#include <X11/keysym.h>
#include <X11/cursorfont.h>
#include <stdlib.h>
#include <stdio.h>
#include <iostream>
#include <string.h>
#include <math.h>
#include <unistd.h>
#include <time.h>
#include <vector>
#include <sys/time.h>
#include <termios.h>
#include "colors.cc"
#include "digits.xbm"
#include "bold-digits.xbm"
#include "small-digits.xbm"
#include "small-bold-digits.xbm"

using namespace std;

Display * dpy;
Window root;
Colormap color;
int screen;
Visual * visual;
uint stopCode;
int macroSlow = 0;
struct termios old_attr;
struct termios new_attr;


typedef void (* EventProcessor)(XEvent *);
struct EventListener {
  static EventListener * listeners;
  EventListener * next;
  EventProcessor process;

  EventListener(EventProcessor p) {
    next = NULL;
    process = p;
  }

  static void init() {
    listeners = NULL;
  }

  static void free() {
    while (listeners != NULL) {
      EventListener * next = listeners->next;
      delete listeners;
      listeners = next;
    }
  }

  static void addListener(EventProcessor p) {
    EventListener * l = new EventListener(p);
    l->next = listeners;
    listeners = l;
  }

  static void removeListener(EventProcessor p) {
    EventListener * l = listeners;
    while (l != NULL) {
      EventListener * next = l->next;
      if (next != NULL && next->process == p) {
	l->next = next->next;
	delete next;
      }
      l = l->next;
    }
  }

  static void processEvent(XEvent * ev) {
    EventListener * l = listeners;
    while (l != NULL) {
      l->process(ev);
      l = l->next;
    }
  }

  static void processEvents() {
    while (XPending(dpy) > 0) {
      XEvent ev;
      XNextEvent(dpy, &ev);
      processEvent(&ev);
    }
  }
};
EventListener * EventListener::listeners;

void printChildren(Window w, int depth) {
  char * windowName;
  XFetchName(dpy, w, &windowName);
  for (int i = 0; i < depth; i++) {
    printf("  ");
  }
  printf("%s\n", windowName ? windowName : "<null>");
  XFree(windowName);

  unsigned int nChildren;
  Window root;
  Window parent;
  Window * children;
  XQueryTree(dpy, w, &root, &parent, &children, &nChildren);
  if (nChildren > 0) {
    for (uint i = 0; i < nChildren; i++) {
      printChildren(children[i], depth + 1);
    }
  }
  XFree(children);
}

int getKey(int timeoutMillis) {
  fd_set rfds;
  FD_ZERO(&rfds);
  FD_SET(0, &rfds);
  struct timeval timeout;
  timeout.tv_sec = timeoutMillis / 1000;
  timeout.tv_usec = (timeoutMillis % 1000) * 1000;
  int readable = select(1, &rfds, 0, 0, &timeout);
  if (!readable)  return -1;
  char c;
  int numChars = read(0, &c, 1);
  if (numChars > 0)  return c;
  return -1;
}

bool getWindowByName(Window base, char * partName, Window * windowReturn) {
  bool found = false;

  char * windowName;
  XFetchName(dpy, base, &windowName);
  if (windowName) {
    found = (strstr(windowName, partName) == windowName);
    if (found) {
      *windowReturn = base;
      //cout << "Found window name: " << windowName << endl;
    }
    XFree(windowName);
    if (found)  return found;
  }

  unsigned int nChildren;
  Window root;
  Window parent;
  Window * children;
  XQueryTree(dpy, base, &root, &parent, &children, &nChildren);
  if (nChildren > 0) {
    for (uint i = 0; i < nChildren && !found; i++) {
      found = getWindowByName(children[i], partName, windowReturn);
    }
  }
  XFree(children);
  return found;
}

ulong getPixel(Window w, int x, int y)
{
  XImage * im = XGetImage(dpy, w, x, y, 1, 1, (unsigned long)-1, ZPixmap);
  ulong pix = XGetPixel(im, 0, 0);
  XDestroyImage(im);
  return pix;
}

void waitForRefresh(Window w)
{
  Window wd;
  int id;
  uint uid;
  uint width;
  uint height;
  XGetGeometry(dpy, w, &wd, &id, &id, &width, &height, &uid, &uid);

  GC gc = XCreateGC(dpy, w, 0, 0);
  ulong WHITE = WhitePixel(dpy, screen);
  XSetForeground(dpy, gc, WHITE);

  int xCheck = width-1;
  int yCheck = 0;
  XDrawPoint(dpy, w, gc, xCheck, yCheck);

  bool refreshed = false;
  while (!refreshed) {
    usleep(20000);
    ulong pixel = getPixel(w, xCheck, yCheck);
    refreshed = (pixel != WHITE);
  }

  XFreeGC(dpy, gc);

  EventListener::processEvents();
}

void button(Window w, int x, int y, int button, bool press) {
  int xOrigin;
  int yOrigin;
  Window wd;
  XTranslateCoordinates(dpy, w, root, 0, 0, &xOrigin, &yOrigin, &wd);
    
  XEvent ev;
  ev.type = (press ? ButtonPress : ButtonRelease);
  ev.xbutton.display = dpy;
  ev.xbutton.window = w;
  ev.xbutton.root = root;
  ev.xbutton.subwindow = w;
  ev.xbutton.time = CurrentTime;
  ev.xbutton.x = x;
  ev.xbutton.y = y;
  ev.xbutton.x_root = x + xOrigin;
  ev.xbutton.y_root = y + yOrigin;
  ev.xbutton.state = (press ? 1 << (button + 7) : 0);
  ev.xbutton.button = button;
  ev.xbutton.same_screen = true;
  XSendEvent(dpy, w, true, 0, &ev);
  if (macroSlow) {
    usleep(macroSlow);
  }
}


void clickAt(Window w, int x, int y) {
  Window root_return, child_return;
  int root_x_return, root_y_return;
  int win_x_return, win_y_return;
  unsigned int mask_return;
 
  Window wd;
  int id;
  uint uid;
  uint width;
  uint height;
  XGetGeometry(dpy, w, &wd, &id, &id, &width, &height, &uid, &uid);

  XQueryPointer(dpy, w, &root_return, &child_return, &root_x_return, &root_y_return, 
		&win_x_return, &win_y_return, &mask_return);
  button(w, win_x_return, win_y_return, Button3, false);
  button(w, win_x_return, win_y_return, Button1, false);
  button(w, x, y, Button3, false);
  //  waitForRefresh(w);
  button(w, x, y, Button3, true);
  button(w, x, y, Button3, false);
  if (win_x_return > 0 &&
      win_y_return > 0 &&
      win_x_return < (int)width &&
      win_y_return < (int)height) {
    button(w, win_x_return, win_y_return, Button3, false);
  }
}

void leftClickAt(Window w, int x, int y) {
  Window root_return, child_return;
  int root_x_return, root_y_return;
  int win_x_return, win_y_return;
  unsigned int mask_return;
  
  Window wd;
  int id;
  uint uid;
  uint width;
  uint height;
  XGetGeometry(dpy, w, &wd, &id, &id, &width, &height, &uid, &uid);

  XQueryPointer(dpy, w, &root_return, &child_return, &root_x_return, &root_y_return, 
		&win_x_return, &win_y_return, &mask_return);
  button(w, win_x_return, win_y_return, Button3, false);
  button(w, win_x_return, win_y_return, Button1, false);
  button(w, x, y, Button1, false);
  waitForRefresh(w);
  button(w, x, y, Button1, true);
  button(w, x, y, Button1, false);
  if (win_x_return > 0 &&
      win_y_return > 0 &&
      win_x_return < (int)width &&
      win_y_return < (int)height) {
    button(w, win_x_return, win_y_return, Button1, false);
  }
}

Color getColor(Window w, int x, int y)
{
  XImage * im = XGetImage(dpy, w, x, y, 1, 1, (unsigned long)-1, ZPixmap);
  Color c = Color::fromImage(im, 0, 0);
  XDestroyImage(im);
  return c;
}

bool matchPixel(Window w, int x, int y, double * match)
{
  return getColor(w, x, y).match(match);
}

double topLine[4] = { 0.93, 0.89, 0.82,   0.20 }; // +- 0.13 total, plus leeway
double bottomLine[4] = { 0.44, 0.40, 0.33,  0.22 }; // +- 0.15 total, plus leeway

bool isOnMenuDivider(Window w, int x, int y) {
  bool match = matchPixel(w, x, y, topLine);
  if (match) {
    match = matchPixel(w, x, y+1, bottomLine);
  }
  return match;
}

struct Range {
  int left;
  int size;
};

struct Point {
  int x;
  int y;
  Point operator+(const Point p2) const {
    return (Point) { x + p2.x, y + p2.y };
  }
  Point operator-(const Point p2) const {
    return (Point) { x - p2.x, y - p2.y };
  }
  Point operator*(const int a) const {
    return (Point) { x*a, y*a };
  }
  Point operator/(const int a) const {
    return (Point) { x/a, y/a };
  }
  bool operator==(const Point p2) const {
    return (x == p2.x && y == p2.y);
  }
  bool veryNear(const Point p2) {
    return ((abs(x - p2.x) < 2 || p2.x < 0) &&
	    (abs(y - p2.y) < 2 || p2.y < 0));
  }
  float length() const {
    return sqrt(x*x + y*y);
  }
};

ostream& operator<<(ostream& os, Point p)
{
  os << "(" << p.x << ", " << p.y << ")";
  return os;
}

XImage * getImage(Window w, Point p, int width, int height)
{
  return XGetImage(dpy, w, p.x, p.y, width, height, (unsigned)-1, ZPixmap);
}

Color getColor(Window w, Point p) {
  XColor xcolor;
  XImage * im = getImage(w, p, 1, 1);
  xcolor.pixel = XGetPixel(im, 0, 0);
  XDestroyImage(im);
  XQueryColor(dpy, color, &xcolor);
  return (Color) { xcolor.red / 65535.0, xcolor.green / 65535.0, xcolor.blue / 65535.0 };
}

bool isPrimary(Color c) {
  return c.isPrimary(0.18);
}

int barLength(Window w, Point p) {
  Point test = p;
  int length = 0;
  while (isPrimary(getColor(w, test))) {
    length++;
    test.x++;
  }
  return length;
}

Range findMenuDivider(Window w, int x, int y) {
  if (!isOnMenuDivider(w, x, y)) {
    return (Range) { x, 0 };
  }
  int left = x - 1;
  while (isOnMenuDivider(w, left, y)) {
    left--;
  }
  left++;
  int right = x + 1;
  while (isOnMenuDivider(w, right, y)) {
    right++;
  }
  right--;
  return (Range) { left, right - left + 1 };
}


struct MacroWindow {
  Window w;
  int xOrigin, yOrigin;
  uint width, height;
};

MacroWindow tale;

XImage * getImage(Point p, int width, int height)
{
  return XGetImage(dpy, tale.w, p.x, p.y, width, height, (unsigned)-1, ZPixmap);
}

Color getPixel(Point p) {
  XImage * im = getImage(p, 1, 1);
  Color c = Color::fromImage(im, 0, 0);
  XDestroyImage(im);
  return c;
}

double strTimerColor[4] = { 0.89, 0.83, 0.70,  0.01 };
Point strTimerOffset = (Point) { -43, 109 };

bool matchPoint(Point p, double * match)
{
  // cout << "Matching at point " << p << endl;
  Color c = getPixel(p);
  return c.match(match);
}

void clickAt(Point p)
{
  clickAt(tale.w, p.x, p.y);
}

void leftClickAt(Point p)
{
  leftClickAt(tale.w, p.x, p.y);
}

void waitForRefresh()
{
  waitForRefresh(tale.w);
}

void clickWait(Point p)
{
  clickAt(p);
  waitForRefresh();
  waitForRefresh();
}

void drag(Point a, Point b)
{
  Window root_return, child_return;
  int root_x_return, root_y_return;
  int win_x_return, win_y_return;
  unsigned int mask_return;
 
  XQueryPointer(dpy, tale.w, &root_return, &child_return, &root_x_return, &root_y_return, 
		&win_x_return, &win_y_return, &mask_return);
  button(tale.w, win_x_return, win_y_return, Button3, false);
  button(tale.w, win_x_return, win_y_return, Button1, false);

  button(tale.w, a.x, a.y, Button3, false);
  waitForRefresh();
  button(tale.w, a.x, a.y, Button3, true);
  button(tale.w, b.x, b.y, Button3, true);
  button(tale.w, b.x, b.y, Button3, false);
  button(tale.w, win_x_return, win_y_return, Button3, false);
  waitForRefresh();
}

long long startTime;

long long _getMillis() {
  struct timeval t;
  gettimeofday(&t, 0);
  return t.tv_sec * 1000 + t.tv_usec / 1000;
}

bool terminalIsRaw = false;

void endMacro()
{
  if (terminalIsRaw) {
    if (isatty(0)) {
      tcsetattr(0, TCSANOW, &old_attr);
    }
  }
  XCloseDisplay(dpy);
  EventListener::free();
}

void processAbort(XEvent * ev)
{
  if (ev->type == KeyPress && ev->xkey.keycode == stopCode) {
    endMacro();
    exit(0);
  }
}

void initMacro() {
  dpy = XOpenDisplay(getenv("DISPLAY"));
  root = XDefaultRootWindow(dpy);
  screen = DefaultScreen(dpy);
  color = DefaultColormap(dpy, screen);
  visual = DefaultVisual(dpy, screen);

  bool found = getWindowByName(root, (char *)"A Tale in the Desert", &tale.w);
  if (!found) {
    printf("Not found\n");
    XCloseDisplay(dpy);
    exit(1);
  }
    
  Window wd;
  XTranslateCoordinates(dpy, tale.w, root, 0, 0, &tale.xOrigin, &tale.yOrigin, &wd);
    
  int id;
  uint uid;
  XGetGeometry(dpy, tale.w, &wd, &id, &id, &tale.width, &tale.height, &uid, &uid);

  stopCode = XKeysymToKeycode(dpy, XK_Pause);
  XGrabKey(dpy, stopCode, AnyModifier, root, GrabModeAsync, GrabModeAsync, CurrentTime);
  startTime = _getMillis();

  EventListener::init();
  EventListener::addListener(processAbort);
}

void makeTerminalRaw()
{
  // Set up terminal
  if (isatty(0)) {
    tcgetattr(0, &old_attr);
    cfmakeraw(&new_attr);
    terminalIsRaw = true;
    tcsetattr(0, TCSANOW, &new_attr);
  }
}


long getMillis()
{
  return (long)(_getMillis() - startTime);
}


char * getTimeString(char * buffer = 0, int length = 0)
{
  // If buffer is null, use static one.
  static char static_buffer[50];
  if (buffer == 0) {
    length = 50;
    buffer = static_buffer;
  }
  time_t now = time(0);
  strftime(buffer, length, "%Y-%m-%d,%H:%M:%S", localtime(&now));
  return buffer;
}

struct BitmapFont {
  unsigned char lowChar;
  int numChars;
  int cellHeight;
  int maxCellWidth;
  unsigned char * data;
  int * cellWidths;

  int getCharAt(Point p) {
    XImage * im = getImage(p, maxCellWidth, cellHeight);
    const static ulong threshold = (im->red_mask / 2) & im->red_mask;

    // Compare with digits
    int digit = 0;
    while (digit < numChars) {
      int matches = 0;
      int cellWidth = cellWidths[digit];
      int match_req = cellHeight * cellWidth - 5;

      for (int i = 0; i < cellHeight; i++) {
	for (int h = cellWidth-1; h >= 0; h--) {
	  ulong pixel = XGetPixel(im, maxCellWidth-1 - h, i);
	  unsigned char target = (data[digit + i * numChars] >> (7-h)) & 0x01;
	  unsigned char bit = 0;
	  if ((pixel & im->red_mask) <= threshold) {
	    bit = 1;
	  }
	  if (bit == target) {
	    matches++;
	  }
	}
      }
      
      if (matches >= match_req) break;
      digit++;
    }

    XDestroyImage(im);

    if (digit < numChars) {
      return digit;
    } else {
      return -1;
    }
    
  }

  int getNumberAt(Point p, int direction = 1)
  {
    int result = 0;
    if (direction < 0) {
	p.x -= maxCellWidth;
    }
    int digit = getCharAt(p);
    if (digit < 0) {
      return -1;
    }
    int multiplier = 1;
    while (digit >= 0) {
      if (direction > 0) {
	result = result * 10 + digit;
	p.x += cellWidths[digit] + 1;
      } else {
	result += digit * multiplier;
	multiplier *= 10;
	p.x -= cellWidths[digit] + 1;
      }
      digit = getCharAt(p);
      if (digit == -1) {
	p.x -= 1;
	digit = getCharAt(p);
      }
    }
    return result;
  }
  
};

int mediumWidths[] = { 6, 6, 6, 6, 6, 6, 6, 6, 6, 6 };
BitmapFont mediumDigits = { '0', 10, digits_height, 6, digits_bits, mediumWidths };

int smallWidths[] = { 5, 4, 5, 5, 5, 5, 5, 5, 5, 5 };
BitmapFont smallDigits = { '0', 10, small_digits_height, 5, small_digits_bits,
			   smallWidths };

int mediumBoldWidths[] = { 7, 7, 7, 7, 7, 7, 7, 7, 7, 7 };
BitmapFont mediumBoldDigits = { '0', 10, bold_digits_height, 7, bold_digits_bits,
				mediumBoldWidths };

int smallBoldWidths[] = { 6, 5, 6, 6, 6, 6, 6, 6, 6, 6 };
BitmapFont smallBoldDigits = { '0', 10, small_bold_digits_height, 6, small_bold_digits_bits,
			       smallBoldWidths };



int getGameTime() {
  // Returns game time within a game month

  // Assumes that a /clockloc window is open at the default location
  Point p = { tale.width/2 + 60, 9 };
  
  while (p.x > (int)tale.width/2 && mediumDigits.getCharAt(p) == -1) {
    p.x--;
  }

  int minutes = mediumDigits.getCharAt(p);
  if (minutes < 0) return -1;

  int min10 = mediumDigits.getCharAt(p - (Point) { 7, 0 });
  if (min10 < 0) return -1;
  minutes += min10 * 10;

  int hour = mediumDigits.getCharAt(p - (Point) { 18, 0 });
  if (hour < 0) return -1;

  int hr10 = mediumDigits.getCharAt(p - (Point) { 25, 0 });
  if (hr10 < 0) {
    hr10 = 0;
  }
  hour += hr10 * 10;

  if (hour == 12) {
    hour = 0;
  }

  if (matchPoint(p + (Point) { 7, 0 }, (double[]) { 0, 0, 0, 0 })) {
    // PM
    hour += 12;
  }

  Point day1Point;
  if (hr10 > 0) {
    day1Point = p - (Point) { 40, 0 };
  } else {
    day1Point = p - (Point) { 33, 0 };
  }
  int day = mediumDigits.getCharAt(day1Point);
  if (day < 0) return -1;

  int day10 = mediumDigits.getCharAt(day1Point - (Point) { 7, 0 });
  if (day10 < 0) {
    day10 = 0;
  }
  day += day10 * 10;

  minutes += hour * 60 + (day - 1) * 1440;

  return minutes;
}

void sendKey(Window w, int x, int y, int key, bool press) {
  int xOrigin;
  int yOrigin;
  Window wd;
  XTranslateCoordinates(dpy, w, root, 0, 0, &xOrigin, &yOrigin, &wd);

  button(w, x, y, 1, 0);

  XEvent ev;
  ev.type = (press ? KeyPress : KeyRelease);
  ev.xkey.display = dpy;
  ev.xkey.window = w;
  ev.xkey.root = root;
  ev.xkey.subwindow = w;
  ev.xkey.time = CurrentTime;
  ev.xkey.x = x;
  ev.xkey.y = y;
  ev.xkey.x_root = x + xOrigin;
  ev.xkey.y_root = y + yOrigin;
  ev.xkey.state = 0;
  ev.xkey.keycode = key;
  ev.xkey.same_screen = true;
  XSendEvent(dpy, w, true, 0, &ev);
  if (macroSlow) {
    usleep(macroSlow);
  }
}

void pressKey(Point p, int keyCode)
{
  sendKey(tale.w, p.x, p.y, keyCode, 1);
  sendKey(tale.w, p.x, p.y, keyCode, 0);
}

void enterString(Point p, const char * string)
{
  char buffer[2];
  buffer[1] = 0;
  while (*string != 0) {
    buffer[0] = *string;
    KeySym ks = XStringToKeysym(buffer);
    KeyCode kc = XKeysymToKeycode(dpy, ks);
    pressKey(p, kc);
    waitForRefresh();
    string++;
  }
}

void enterNumber(int number)
{
  Point p = {0, 0};
  char numberString[12];
  snprintf(numberString, sizeof(numberString), "%d", number);
  enterString(p, numberString);
  pressKey(p, XKeysymToKeycode(dpy, XK_Return));
}

struct Rectangle {
  Point topLeft;
  Point size;
};
typedef vector<Rectangle *> RectList;


double blackPixel[] = { 0, 0, 0, 0.05 };
double whiteBorder[] = { .98, .97, .95, .09 };
double brownBorder[] = { .57, .42, .27, .15 };
double blackBorder[] = { .13, .12, .10, .09 };

/*
Old Tale 1 - 3 menus

Rectangle * getMenu(XImage * im, int x, int y, bool checkThis = true) {
  if (checkThis) {
    // Look for a "T" character inside white/black borders at least 64x64
    // Verify that "T" shape is black, and outside the T is not black
    for (int dx = 10; dx < 16; dx++) {
      if (XGetPixel(im, x+dx, y+13) == 0) return 0;
      if (XGetPixel(im, x+dx, y+14) != 0) return 0;
    }
    for (int dy = 15; dy < 23; dy++) {
      if (XGetPixel(im, x+11, y+dy) == 0) return 0;
      if (XGetPixel(im, x+12, y+dy) != 0) return 0;
      if (XGetPixel(im, x+13, y+dy) != 0) return 0;
      if (XGetPixel(im, x+14, y+dy) == 0) return 0;
    }
  }
  // Find white edges of menu
  uint borderX = x;
  while (borderX < tale.width && Color::fromImage(im, borderX, y).match(whiteBorder)) {
    borderX++;
  }
  if (borderX - x < 60 || borderX >= tale.width) return 0;

  uint borderY = y;
  while (borderY < tale.height && Color::fromImage(im, x, borderY).match(whiteBorder)) {
    borderY++;
  }
  if (borderY - y < 60 || borderY >= tale.height) return 0;
  
  // Check that black edges are where they ought to be
  for (uint x1 = x; x1 < borderX; x1++) {
    if (!Color::fromImage(im, x1, borderY).match(blackBorder)) return 0;
  }
  for (uint y1 = y; y1 < borderY; y1++) {
    if (!Color::fromImage(im, borderX, y1).match(blackBorder)) return 0;
  }

  Rectangle * r = new Rectangle();
  r->topLeft.x = x;
  r->topLeft.y = y;
  r->size.x = borderX + 1 - x;
  r->size.y = borderY + 1 - y;
  return r;
}
*/

Rectangle * getMenuAtImpl(XImage * im, int x, int y, int checkThis = false)
{
  if (checkThis) {
    // Look for a "T" character inside borders at least 64x64
    // Verify that "T" shape is black, and outside the T is not black
    for (int dx = 10; dx < 16; dx++) {
      if (XGetPixel(im, x+dx, y+13) == 0) return 0;
      if (XGetPixel(im, x+dx, y+14) != 0) return 0;
    }
    for (int dy = 15; dy < 23; dy++) {
      if (XGetPixel(im, x+11, y+dy) == 0) return 0;
      if (XGetPixel(im, x+12, y+dy) != 0) return 0;
      if (XGetPixel(im, x+13, y+dy) != 0) return 0;
      if (XGetPixel(im, x+14, y+dy) == 0) return 0;
    }
    // Now move into the menu interior and check the rest
    x += 10;
    y += 10;
  }

  // Find inner left brown border
  while (!Color::fromImage(im, x, y).match(brownBorder)) {
    x--;
    if (x < 2) return 0;
  }

  // Find inner top black border
  while (Color::fromImage(im, x, y).match(brownBorder)) {
    y--;
    if (y < 3) return 0;
  }
  if (!Color::fromImage(im, x, y).match(blackPixel)) return 0;
  x++;

  // Trace black border, looking for right brown border underneath
  int right = x;
  while (!Color::fromImage(im, right, y+1).match(brownBorder)) {
    if (!Color::fromImage(im, right, y).match(blackPixel)) return 0;
    right++;
    if (right + 2 >= im->width) return 0;
  }

  // Find black border on the bottom
  int bottom = y + 1;
  while (Color::fromImage(im, right, bottom).match(brownBorder)) {
    bottom++;
    if (bottom + 2 >= im->height) return 0;
  }
  if (!Color::fromImage(im, right, bottom).match(blackPixel)) return 0;
  right--;

  // Reject regions that are too small
  if (right < x + 40 || bottom < y + 40) return 0;

  x -= 3;
  y -= 3;
  right += 4;
  bottom += 4;

  Rectangle * menu = new Rectangle();
  menu->topLeft.x = x;
  menu->topLeft.y = y;
  menu->size.x = right - x;
  menu->size.y = bottom - y;
  //cout << "Menu at " << menu->topLeft << "-" << (menu->topLeft + menu->size) << endl;
  return menu;
}

Rectangle * getMenuAt(Point p)
{
  XImage * im = getImage((Point){0, 0}, tale.width, tale.height);
  Rectangle * result = getMenuAtImpl(im, p.x, p.y);
  XDestroyImage(im);
  return result;
}

Rectangle * getMenuAtTry(Point p)
{
  Rectangle * menu = getMenuAt(p);
  while (!menu) {
    waitForRefresh();
    menu = getMenuAt(p);
  }
  return menu;
}


RectList * getMenus(bool checkThis = true)
{
  RectList * list = new RectList();
  XImage * im = getImage((Point){0, 0}, tale.width, tale.height);
  for (uint y = 0; y < tale.height - 40; y++) {
    for (uint x = 0; x < tale.width - 40; x++) {
      Rectangle * r = getMenuAtImpl(im, x, y, checkThis);
      if (r) {
	list->push_back(r);
	x += r->size.x;
      }
    }
  }
  XDestroyImage(im);
  return list;
}

bool getPointer(Point & p, uint cursorShape = XC_target) {
  Cursor cursor = XCreateFontCursor(dpy, cursorShape);

  XSetWindowAttributes swa;
  swa.event_mask = ButtonPressMask;
  swa.cursor = cursor;

  Window w = XCreateWindow(dpy, tale.w,
			   0, 0,
			   tale.width, tale.height,
			   0, 0, InputOnly,
			   0, CWCursor | CWEventMask, &swa);
  XMapWindow(dpy, w);

  bool gotPointer = false;
  bool isValid = false;
  while (!gotPointer) {
    XEvent ev;
    XNextEvent(dpy, &ev);
    if (ev.type == ButtonPress) {
      p.x = ev.xbutton.x;
      p.y = ev.xbutton.y;
      isValid = (ev.xbutton.button == Button1);
      gotPointer = true;
    }
  }

  XDestroyWindow(dpy, w);
  XFreeCursor(dpy, cursor);
  //cout << "Clicked at " << p << (isValid ? "" : " - cancelled") << endl;
  return isValid;
}


bool queryPointer(Point & p) {
  Window child, root;
  int x, y, root_x, root_y;
  uint mask;
  Bool b = XQueryPointer(dpy, tale.w, &root, &child, &root_x, &root_y, &x, &y, &mask);
  p.x = x;
  p.y = y;
  return (b == True && x >= 0 && y >= 0 && x < (int)tale.width && y < (int)tale.height);
}
